@@ -1,12 +1,12 @@
 /* pam_warn module */
 
 /*
- * $Id: pam_warn.c,v 1.3 2005/06/09 17:29:20 kukuk Exp $
+ * $Id$
  *
  * Written by Andrew Morgan <morgan@linux.kernel.org> 1996/3/11
  */
 
-#define _BSD_SOURCE
+#include "config.h"
 
 #include <stdio.h>
 #include <unistd.h>
@@ -24,6 +24,7 @@
 #define PAM_SM_PASSWORD
 
 #include <security/pam_modules.h>
+#include <security/pam_ext.h>
 
 /* some syslogging */
 
@@ -32,18 +33,7 @@
      value = value ? value : default_value ;                    \
 } while (0)
 
-static void _pam_log(int err, const char *format, ...)
-{
-    va_list args;
-
-    va_start(args, format);
-    openlog("PAM-warn", LOG_CONS|LOG_PID, LOG_AUTH);
-    vsyslog(err, format, args);
-    va_end(args);
-    closelog();
-}
-
-static void log_items(pam_handle_t *pamh, const char *function)
+static void log_items(pam_handle_t *pamh, const char *function, int flags)
 {
      const void *service=NULL, *user=NULL, *terminal=NULL,
 	 *rhost=NULL, *ruser=NULL;
@@ -54,57 +44,63 @@ static void log_items(pam_handle_t *pamh, const char *function)
      OBTAIN(PAM_RUSER, ruser, "<unknown>");
      OBTAIN(PAM_RHOST, rhost, "<unknown>");
 
-     _pam_log(LOG_NOTICE, "function=[%s] service=[%s] terminal=[%s] user=[%s]"
-	      " ruser=[%s] rhost=[%s]\n",
-	      function, service, terminal, user, ruser, rhost);
+     pam_syslog(pamh, LOG_NOTICE,
+		"function=[%s] flags=%#x service=[%s] terminal=[%s] user=[%s]"
+		" ruser=[%s] rhost=[%s]\n", function, flags,
+		(const char *) service, (const char *) terminal,
+		(const char *) user, (const char *) ruser,
+		(const char *) rhost);
 }
 
 /* --- authentication management functions (only) --- */
 
 PAM_EXTERN
-int pam_sm_authenticate(pam_handle_t *pamh, int flags, int argc,
-			const char **argv)
+int pam_sm_authenticate(pam_handle_t *pamh, int flags,
+			int argc UNUSED, const char **argv UNUSED)
 {
-    log_items(pamh, __FUNCTION__);
+    log_items(pamh, __FUNCTION__, flags);
     return PAM_IGNORE;
 }
 
 PAM_EXTERN
-int pam_sm_setcred(pam_handle_t *pamh, int flags, int argc, const char **argv)
+int pam_sm_setcred(pam_handle_t *pamh, int flags,
+		   int argc UNUSED, const char **argv UNUSED)
 {
-    log_items(pamh, __FUNCTION__);
+    log_items(pamh, __FUNCTION__, flags);
     return PAM_IGNORE;
 }
 
 /* password updating functions */
 
 PAM_EXTERN
-int pam_sm_chauthtok(pam_handle_t *pamh,int flags,int argc,const char **argv)
+int pam_sm_chauthtok(pam_handle_t *pamh, int flags,
+		     int argc UNUSED, const char **argv UNUSED)
 {
-    log_items(pamh, __FUNCTION__);
+    log_items(pamh, __FUNCTION__, flags);
     return PAM_IGNORE;
 }
 
 PAM_EXTERN int
-pam_sm_acct_mgmt (pam_handle_t *pamh, int flags, int argc, const char **argv)
+pam_sm_acct_mgmt(pam_handle_t *pamh, int flags,
+		 int argc UNUSED, const char **argv UNUSED)
 {
-    log_items(pamh, __FUNCTION__);
+    log_items(pamh, __FUNCTION__, flags);
     return PAM_IGNORE;
 }
 
 PAM_EXTERN int
-pam_sm_open_session (pam_handle_t *pamh, int flags, int argc,
-                     const char **argv)
+pam_sm_open_session(pam_handle_t *pamh, int flags,
+		    int argc UNUSED, const char **argv UNUSED)
 {
-    log_items(pamh, __FUNCTION__);
+    log_items(pamh, __FUNCTION__, flags);
     return PAM_IGNORE;
 }
 
 PAM_EXTERN int
-pam_sm_close_session (pam_handle_t *pamh, int flags, int argc,
-		      const char **argv)
+pam_sm_close_session(pam_handle_t *pamh, int flags,
+		     int argc UNUSED, const char **argv UNUSED)
 {
-    log_items(pamh, __FUNCTION__);
+    log_items(pamh, __FUNCTION__, flags);
     return PAM_IGNORE;
 }
 
